unit uPlot_ResultWriter;
{
*****************************************************************************
*  This file is part of Multiple Acronym Math and Audio Plot - MAAPlot      *
*                                                                           *
*  See the file COPYING.                                                    *
*  for details about the copyright.                                         *
*                                                                           *
*  This program is distributed in the hope that it will be useful,          *
*  but WITHOUT ANY WARRANTY; without even the implied warranty of           *
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                     *
*                                                                           *
*  MAAplot for Lazarus/fpc                                                  *
*  (C) 2014 Stefan Junghans                                                 *
*****************************************************************************
}
{
Export / Import handler for MAAplot
implemented:
  TPlot_ResultWriterMAAResult
    writes one or more MAAResult (see Dataytpes, TMeasResult)
    into a filestream. Adds a meaningful header before the Datasets.
    Data is binary, structure see this class.

  TPlot_ResultReaderMAAResult
    reads Data written by the TPlot_ResultWriterMAAResult

  TPlot_ResultReaderCSV
    reads a CSV file with help from < csvdocument > class fpr Lazarus/fpc
    see < CSV Parser, Builder and Document classes > for copyright, authors and license
    Note: assumption need to be made about CSV file structiore which is not specified:
    - Specific number of headerlines can be skipped
    - Datapoints are in rows (one row, one datapoint)
    - First column is X value
    - Further columns are
        a) furter dimensions (Y, Z, ....).
        b) treated as separate Y values all with the same X value in column 0
}

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, dateutils, uPlotDataTypes,
  CsvDocument, uPlot_ResultWriterHelperForms;

type

  TExportResultType = (ertMAAResult);

  PExportHeader = ^TExportHeader;
  TExportHeader = record
    dwNumResults: Integer;
    dwDate:       TDateTime;
    dwDescription: ShortString;
    dwReserved: ShortString;
    dwVersion: Integer;
    dwResultType: TExportResultType;
  end;

  PExportResult_MaaResult = ^TExportResult_MAAResult;
  TExportResult_MAAResult = record
    dwHeader: TExportHeader;
    dwResultArray: array of TMeasResult;
  end;

  { TPlot_ResultWriter }

  TPlot_ResultWriter = class
  private
    function GetDataDateTime: TDateTime;
    function GetDataDescription: ShortString;
    procedure SetDataDateTime(AValue: TDateTime);
    procedure SetDataDescription(AValue: ShortString);
  protected
    FDataType: TExportResultType;
    FDataHeader: TExportHeader;
  public
    constructor Create; virtual;
    destructor Destroy; override;

    function WriteToFile(AFileName: TFilename): Integer; virtual;abstract;
    procedure ReSet;virtual;

    property DataDescription: ShortString read GetDataDescription write SetDataDescription;
    property DataDateTime: TDateTime read GetDataDateTime write SetDataDateTime;
  end;

  { TPlot_ResultWriterMAAResult }

  TPlot_ResultWriterMAAResult = class (TPlot_ResultWriter)
  private
    FData: TExportResult_MAAResult;
  protected
    function PrepareWrite: Integer;
  public
    constructor Create; override;
    destructor Destroy; override;
    procedure ReSet;override;

    function WriteToFile(AFileName: TFilename): Integer; override;
    function AddResultRecord(AResultRecord: TMeasResult): Integer;
  end;

  { TPlot_ResultReader }

  TPlot_ResultReader = class
  private
    function GetDataDateTime: TDateTime;
    function GetDataDescription: ShortString;
    function GetNumResults: Integer;
    function GetResultType: TExportResultType;
  protected
    FDataHeader: TExportHeader;
    FData : TExportResult_MAAResult;
  public
    constructor Create; virtual;
    destructor Destroy; override;

    function LoadFromFile(AFileName: TFilename): Integer; virtual;abstract;
    procedure ReSet;virtual;

    function GetResultRecord(AResultIndex: Integer; var AResultRecord: TMeasResult): Integer;virtual;

    property DataDescription: ShortString read GetDataDescription;
    property DataDateTime: TDateTime read GetDataDateTime;
    property ResultType: TExportResultType read GetResultType;
    property NumResults: Integer read GetNumResults;

  end;

  { TPlot_ResultReaderMAAResult }

  TPlot_ResultReaderMAAResult = class (TPlot_ResultReader)
  private
    //FData : TExportResult_MAAResult;
  protected
  public
    constructor Create; override;
    destructor Destroy; override;

    function LoadFromFile(AFileName: TFilename): Integer; override;
    procedure ReSet;override;

    // not dll safe, use only within plot (there is no data handshake. datapointer in handeled inside this funtion)
    function GetResultRecord(AResultIndex: Integer; var AResultRecord: TMeasResult): Integer;override;

  end;

  { TPlot_ResultReaderCSV }

  TPlot_ResultReaderCSV = class (TPlot_ResultReader)
  private
    FHelperForm: THelperFormResultRW;
    //FData : TExportResult_MAAResult;
    FCSVDocument: TCSVDocument;
    FCSVDelimiter: Char;
    FCSVNumHeaderLines: Integer;
    FCSVVariant: TCSVVariant;
  protected
    procedure OnHelperFormDone(Sender: TObject);
  public
    constructor Create; override;
    destructor Destroy; override;

    function LoadFromFile(AFileName: TFilename): Integer; override;
    procedure ReSet;override;

    // not dll safe, use only within plot (there is no data handshake. datapointer in handeled inside this funtion)
    function GetResultRecord(AResultIndex: Integer; var AResultRecord: TMeasResult): Integer;override;
  end;


const
  cINVALIDVALUE = -1e999;


implementation

{ TPlot_ResultReaderCSV }

procedure TPlot_ResultReaderCSV.OnHelperFormDone(Sender: TObject);
begin
  // update the CSV options when helperform closes
  FCSVDelimiter := THelperFormResultReaderCSV(FHelperForm).CSVDelimiter;
  FCSVNumHeaderLines := THelperFormResultReaderCSV(FHelperForm).CSVNumHeaderLines;
  FCSVVariant := THelperFormResultReaderCSV(FHelperForm).CSVvariant;
end;

constructor TPlot_ResultReaderCSV.Create;
begin
  inherited Create;
  FCSVNumHeaderLines := 0;
  FCSVDelimiter := ',';
  FCSVDocument := TCSVDocument.Create;
  FCSVVariant := csvvSingleSet;
  FHelperForm := THelperFormResultReaderCSV.Create(nil);
  FHelperForm.DoneCallbackProc := @OnHelperFormDone;
end;

destructor TPlot_ResultReaderCSV.Destroy;
begin
  FHelperForm.Free;
  FCSVDocument.Free;
  inherited Destroy;
end;

function TPlot_ResultReaderCSV.LoadFromFile(AFileName: TFilename): Integer;
var
  vRowCount, vColCount: Integer;
  vHeaderItem: TExportHeader;
  vIdentItem: TMeasResultIdent;
  vLoop: Integer;
  vLoopDim, vLoopPt: Integer;
  vNumResults, vDimensions, vPoints: Integer;
  vActualDim, vActualPt: Integer;
  vFormatSettings: TFormatSettings;
  vValueStr: String;
begin
  vFormatSettings.DecimalSeparator := '.';
  vFormatSettings.ThousandSeparator := char('');

  // show a helper form for import options
  THelperFormResultReaderCSV(FHelperForm).PreView(AFileName);
  FHelperForm.ShowModal;

  FCSVDocument.Delimiter := FCSVDelimiter;
  //FCSVDocument.QuoteChar := '"';  // do we need this ?
  FCSVDocument.LoadFromFile(AFileName);

  vRowCount := FCSVDocument.RowCount;
  vColCount := FCSVDocument.ColCount[0];
  //writeln('rows: ', IntToStr(vRowCount), '   cols: ', IntToStr(vColCount));

  vPoints := vRowCount - FCSVNumHeaderLines;  // import all rows points except numheaderlines
  if (vPoints < 1) or (vColCount < 2) then begin
    Result := 0;
    exit;
  end;
  case FCSVVariant of
    csvvSingleSet: begin
                     vDimensions := vColCount;      // each column represetnts one dimension
                     vNumResults := 1;
                   end;
    csvvXMultiY:   begin
                     vDimensions := 2;              // first column is always X, further olumns are Y values for that same X
                     vNumResults := vColCount-1;
                   end;
  end;

  // 1st col is X value ?
  // other cols are all Y values ?
  // or dimensions ?
  // --> if all other are Y values, we have mutiple datasets
  // for storage in FData
  // otherwise only one set with n dimensions

  with vHeaderItem do begin // init for safety only
    dwNumResults := vNumResults;
    dwDate := Today;
    dwDescription := 'none (CSV imported)';
    dwReserved :=  '';
    dwVersion := 0;
    dwResultType := ertMAAResult;
  end;

  with vIdentItem do begin
    dwResultType:= mrtGenResult;
    diDimensions:= vDimensions;
    diPoints:= vPoints;
    diMeasurement:= 0;
    dwTimeStamp:= DateTimeToTimeStamp(Now);
  end;

  // clear Data
  for vLoop := 0 to length(FData.dwResultArray)-1 do begin
    setlength(FData.dwResultArray[vLoop].lpMeasData^, 0, 0);
    Dispose(FData.dwResultArray[vLoop].lpMeasData);
  end;
  setlength(FData.dwResultArray, vNumResults);

  // copy CELLS form FCSVDocument to MAAresult
  writeln('results: ', IntToStr(vNumResults), '   dimensions: ', IntToStr(vDimensions), '   points: ', IntToStr(vPoints));

  for vLoop := 0 to vNumResults-1 do begin
    FData.dwResultArray[vLoop].dwResultIdent := vIdentItem;
    FDataHeader := vHeaderItem;
    FData.dwHeader := vHeaderItem;
    New(FData.dwResultArray[vLoop].lpMeasData);
    try
      setlength( FData.dwResultArray[vLoop].lpMeasData^, vPoints, vDimensions );

      for vLoopPt := 0 to vPoints-1 do begin
        vActualPt := vLoopPt;
        case FCSVVariant of
          csvvSingleSet: begin
                           for vLoopDim := 0 to vDimensions-1 do begin
                             vValueStr := FCSVDocument.Cells[vLoopDim, vActualPt+FCSVNumHeaderLines];
                             FData.dwResultArray[vLoop].lpMeasData^[vActualPt, vLoopDim] :=
                               StrToFloatDef(vValueStr, cINVALIDVALUE, vFormatSettings);
                             writeln('dim ', IntToStr(vLoopDim),' string>',FCSVDocument.Cells[vLoopDim, vActualPt+FCSVNumHeaderLines],'<');
                             writeln('dim ', IntToStr(vLoopDim),' string>',vValueStr,'<');
                           end;
                         end;
          csvvXMultiY:   begin
                           vActualDim := vLoop+1;
                           FData.dwResultArray[vLoop].lpMeasData^[vActualPt, 1] :=
                             StrToFloatDef(FCSVDocument.Cells[vActualDim, vActualPt+FCSVNumHeaderLines], cINVALIDVALUE, vFormatSettings);
                           FData.dwResultArray[vLoop].lpMeasData^[vActualPt, 0] :=
                             StrToFloatDef(FCSVDocument.Cells[0, vActualPt+FCSVNumHeaderLines], cINVALIDVALUE, vFormatSettings);
                         end;
        end;
      end;

    except
      setlength(FData.dwResultArray[vLoop].lpMeasData^, 0, 0);
      Dispose(FData.dwResultArray[vLoop].lpMeasData);
    end;
  end;

  Result := vNumResults;
end;

procedure TPlot_ResultReaderCSV.ReSet;
begin
  inherited ReSet;
end;

function TPlot_ResultReaderCSV.GetResultRecord(AResultIndex: Integer;
  var AResultRecord: TMeasResult): Integer;
begin
  Result := inherited GetResultRecord(AResultIndex, AResultRecord);
end;

{ TPlot_ResultReaderMAAResult }

constructor TPlot_ResultReaderMAAResult.Create;
begin
  inherited Create;
end;

destructor TPlot_ResultReaderMAAResult.Destroy;
begin
  inherited Destroy;
end;

function TPlot_ResultReaderMAAResult.LoadFromFile(AFileName: TFilename
  ): Integer;
var
  vStream: TFileStream;
  vFileName: TFilename;
  vLoop, vLoopRow, vLoopDim: Integer;
  vHeaderItem: TExportHeader;
  vDataItem: PMeasArray;
  vIdentItem: TMeasResultIdent;
  vReadCount: Integer;
  vNumResults: Integer;
begin
  Result := 0;
  vFileName := AFileName;
  vReadCount:=0;

  vStream := nil;

  with vHeaderItem do begin // init for safety only
    dwNumResults := 0;
    dwDate := Today;
    dwDescription := 'empty';
    dwReserved :=  '';
    dwVersion := 0;
    dwResultType := ertMAAResult;
  end;

  with vIdentItem do begin
    dwResultType:= mrtGenResult;
    diDimensions:= 0;
    diPoints:= 0;
    diMeasurement:= 0;
    dwTimeStamp:= DateTimeToTimeStamp(Now);
  end;


  // clear Data
  for vLoop := 0 to length(FData.dwResultArray)-1 do begin
    setlength(FData.dwResultArray[vLoop].lpMeasData^, 0, 0);
    Dispose(FData.dwResultArray[vLoop].lpMeasData);
  end;
  setlength(FData.dwResultArray, 0);


  try
    if FileExists(vFileName) then
      vStream := TFileStream.Create(vFileName, fmOpenRead)
    else vStream := nil;
  except vStream := nil; end;

  if vStream = nil then exit;

  try
    // cannot check this as items have different sizes ;-)
    //try
    //  if (vStream.Size MOD vItemSize) <>0 then
    //    raise Exception.Create('stored plotdata is corrupted ..');
    //except
    //  Application.HandleException(Self);
    //  vStream.Free;
    //  vStream := TFileStream.Create(vFileName, fmCreate or fmOpenWrite);
    //  vStream.Size := 0;
    //end;

    // load items
    vStream.Position := 0;

    // load header
    vStream.Read(vHeaderItem, SizeOf(TExportHeader));
    vReadCount := vReadCount + SizeOf(TExportHeader);
    FDataHeader := vHeaderItem;

    vNumResults := vHeaderItem.dwNumResults;
    // prepare FData strructire
    setlength(FData.dwResultArray, vNumResults);
    for vLoop := 0 to vNumResults-1 do begin

      // read ident
       vStream.Read(vIdentItem, SizeOf(TMeasResultIdent));
       vReadCount := vReadCount + SizeOf(TMeasResultIdent);
       // prepare FData structire
       New(vDataItem);
       try
       SetLength(vDataItem^, vIdentItem.diPoints, vIdentItem.diDimensions);
       // read data
       for vLoopRow :=0 to vIdentItem.diPoints-1 do
         for vLoopDim :=0 to vIdentItem.diDimensions-1 do
           vStream.Read(vDataItem^[vLoopRow, vLoopDim], SizeOf(TValue));;

       // store
       FData.dwResultArray[vLoop].dwResultIdent := vIdentItem;
       FData.dwResultArray[vLoop].lpMeasData := vDataItem;

       FData.dwHeader := FDataHeader;
       except
         SetLength(vDataItem^, 0, 0);
         Dispose(vDataItem);
       end;
    end;
  finally
    {$IFDEF DEBUG}OutputDebugString(PChar('Headers loaded bytes/elements: ' + IntToStr(vStream.Size) + ' / ' + IntToStr(Result) ));{$ENDIF}
    FreeAndNil(vStream);
  end;

  Result := vNumResults; // vReadCount;
end;

procedure TPlot_ResultReaderMAAResult.ReSet;
begin
  inherited ReSet;
end;

function TPlot_ResultReaderMAAResult.GetResultRecord(AResultIndex: Integer;
  var AResultRecord: TMeasResult): Integer;
begin
  Result := inherited GetResultRecord(AResultIndex, AResultRecord);
end;

{ TPlot_ResultReader } //=======================================================

function TPlot_ResultReader.GetDataDateTime: TDateTime;
begin
  Result := FDataHeader.dwDate;
end;

function TPlot_ResultReader.GetDataDescription: ShortString;
begin
  Result := FDataHeader.dwDescription;
end;

function TPlot_ResultReader.GetNumResults: Integer;
begin
  Result := FDataHeader.dwNumResults;
end;

function TPlot_ResultReader.GetResultType: TExportResultType;
begin
  Result := FDataHeader.dwResultType;
end;

constructor TPlot_ResultReader.Create;
begin
  inherited Create;
  ReSet;
end;

destructor TPlot_ResultReader.Destroy;
var
  vLoop: Integer;
begin
  IF length(FData.dwResultArray) > 0 then begin

    for vLoop := 0 to length(FData.dwResultArray)-1 do begin
      setlength(FData.dwResultArray[vLoop].lpMeasData^, 0, 0);
      Dispose(FData.dwResultArray[vLoop].lpMeasData);
    end;
    setlength(FData.dwResultArray, 0);

  end;
  ReSet;
  inherited Destroy;
end;

procedure TPlot_ResultReader.ReSet;
begin
  with FDataHeader do begin
    dwNumResults := 0;
    dwDescription := 'nothing loaded';
    dwDate := Now;
    dwReserved := '';
    dwVersion := -1;
    dwResultType := ertMAAResult;
  end;
end;

function TPlot_ResultReader.GetResultRecord(AResultIndex: Integer;
  var AResultRecord: TMeasResult): Integer;
var
  vLoopDim, vLoopRow: Integer; // vLoop
begin
  Result := 0;
  IF AResultIndex > (length(FData.dwResultArray) -1) THEN begin
    Result := -1;
    exit;
  END;
  //New(vMeasResult.lpMeasData); // already created in caller
  try
    SetLength(AResultRecord.lpMeasData^, FData.dwResultArray[AResultIndex].dwResultIdent.diPoints, FData.dwResultArray[AResultIndex].dwResultIdent.diDimensions);
    AResultRecord.dwResultIdent := FData.dwResultArray[AResultIndex].dwResultIdent;
    for vLoopRow :=0 to FData.dwResultArray[AResultIndex].dwResultIdent.diPoints-1 do
      for vLoopDim :=0 to FData.dwResultArray[AResultIndex].dwResultIdent.diDimensions-1 do
        AResultRecord.lpMeasData^[vLoopRow, vLoopDim] := FData.dwResultArray[AResultIndex].lpMeasData^[vLoopRow, vLoopDim];

    AResultRecord.dwResultIdent := FData.dwResultArray[AResultIndex].dwResultIdent;
  except
    setlength(AResultRecord.lpMeasData^, 0, 0);
    Dispose(AResultRecord.lpMeasData);
    Result := -1;
  end;
end;

{ TPlot_ResultWriterMAAResult }  //=============================================

function TPlot_ResultWriterMAAResult.PrepareWrite: Integer;
begin
  FDataHeader.dwNumResults := length(FData.dwResultArray);
  FData.dwHeader := FDataHeader;
  Result := FDataHeader.dwNumResults;
end;

constructor TPlot_ResultWriterMAAResult.Create;
begin
  inherited Create;
end;

destructor TPlot_ResultWriterMAAResult.Destroy;
begin
  inherited Destroy;
end;

procedure TPlot_ResultWriterMAAResult.ReSet;
var
  vLoop: Integer;
begin
  for vLoop := 0 to length(FData.dwResultArray)-1 do begin
    setlength(FData.dwResultArray[vLoop].lpMeasData^, 0, 0);
    Dispose(FData.dwResultArray[vLoop].lpMeasData);
  end;
  setlength(FData.dwResultArray, 0);
  //
  with FDataHeader do begin
    dwNumResults := 0;
    dwDescription := 'MAA export';
    dwDate := Now;
    dwReserved := '2secnutw3tchndagckluehlasti87l339p6czn93vnrelc83o52p0z7'; // write a sync structure, costs nothing, might help in the future
    dwVersion := 0;
    dwResultType := ertMAAResult;
  end;
end;

function TPlot_ResultWriterMAAResult.WriteToFile(AFileName: TFilename): Integer;
var
  vStream: TFileStream;
  vLoop, vLoopDim, vLoopRow: Integer;
  vFileName: TFilename;
  vHeaderSize: Integer;
  vIdentSize, vRowSize: Integer;
  vWriteCount: Integer;
begin
  // writes as follows:
  // <GlobalHeader><ResultIdent><ResultData><ResultIdent><ResultData><ResultIdent><ResultData>  ....
  vFileName:=AFileName;
  Result := 0;
  vWriteCount:=0;
  //vFileName := FMsgPath + cPATHDELIMITER + cHEADERFILENAME + cCONTENTEXTENSION;

  PrepareWrite;

  vStream := TFileStream.Create(vFileName, fmCreate);
  vStream.Position := 0;    // TODO: not needed
  try
    // write Header
    vHeaderSize := sizeof(TExportHeader);
    vStream.Write(FData.dwHeader, vHeaderSize);
    vWriteCount := vWriteCount + vHeaderSize;

    vIdentSize := SizeOf(TMeasResultIdent);

    for vLoop := 0 to FData.dwHeader.dwNumResults-1 do begin
      //vValueCount := FData.dwResultArray[vLoop].dwResultIdent.diPoints * FData.dwResultArray[vLoop].dwResultIdent.diDimensions;
      vRowSize := FData.dwResultArray[vLoop].dwResultIdent.diPoints;
      // write ResultIdent
      vStream.Write(FData.dwResultArray[vLoop].dwResultIdent, vIdentSize);
      // write Data
      for vLoopRow := 0 to vRowSize-1 do
        for vLoopDim := 0 to FData.dwResultArray[vLoop].dwResultIdent.diDimensions-1 do
          vStream.Write(FData.dwResultArray[vLoop].lpMeasData^[vLoopRow, vLoopDim], SizeOf(TValue));

      vWriteCount := vWriteCount + vIdentSize + vRowSize * FData.dwResultArray[vLoop].dwResultIdent.diDimensions * SizeOf(TValue);
    end;
    Result := vWriteCount;
  finally
    FreeAndNil(vStream);
  end;
  ReSet;
end;

function TPlot_ResultWriterMAAResult.AddResultRecord(AResultRecord: TMeasResult
  ): Integer;
var
  vLoopDim, vLoopRow: Integer;  // vLoop
  vMeasResult: TMeasResult;
  //vValueCount, vItemSize: Integer;
begin
  Result := 0;

  IF (AResultRecord.dwResultIdent.diPoints < 1) OR (AResultRecord.dwResultIdent.diDimensions < 1) THEN begin
    Result := -1;
    exit;
  END;

  SetLength(FData.dwResultArray, length(FData.dwResultArray) + 1);

  vMeasResult.dwResultIdent := AResultRecord.dwResultIdent;
  New(vMeasResult.lpMeasData);
  setlength(vMeasResult.lpMeasData^, AResultRecord.dwResultIdent.diPoints, AResultRecord.dwResultIdent.diDimensions);
  try
    vMeasResult.dwResultIdent := AResultRecord.dwResultIdent;
    for vLoopRow :=0 to vMeasResult.dwResultIdent.diPoints-1 do
      for vLoopDim :=0 to vMeasResult.dwResultIdent.diDimensions-1 do
        vMeasResult.lpMeasData^[vLoopRow, vLoopDim] := AResultRecord.lpMeasData^[vLoopRow, vLoopDim];


    //vValueCount := vMeasResult.dwResultIdent.diPoints * vMeasResult.dwResultIdent.diDimensions;
    //vItemSize := SizeOf(TMeasResultIdent) + vValueCount * SizeOf(TValue);

    FData.dwResultArray[length(FData.dwResultArray)-1] := vMeasResult;
  except
    setlength(vMeasResult.lpMeasData^, 0, 0);
    Dispose(vMeasResult.lpMeasData);
  end;

end;

{ TPlot_ResultWriter } //=======================================================

function TPlot_ResultWriter.GetDataDateTime: TDateTime;
begin
  Result := FDataHeader.dwDate;
end;

function TPlot_ResultWriter.GetDataDescription: ShortString;
begin
  Result := FDataHeader.dwDescription;
end;

procedure TPlot_ResultWriter.SetDataDateTime(AValue: TDateTime);
begin
  FDataHeader.dwDate := AValue;
end;

procedure TPlot_ResultWriter.SetDataDescription(AValue: ShortString);
begin
  FDataHeader.dwDescription := AValue;
end;

constructor TPlot_ResultWriter.Create;
begin
  inherited Create;
  ReSet;
end;

destructor TPlot_ResultWriter.Destroy;
begin
  ReSet;
  inherited Destroy;
end;

procedure TPlot_ResultWriter.ReSet;
begin
  with FDataHeader do begin
    dwNumResults := 0;
    dwDescription := 'MAA export';
    dwDate := Now;
    dwReserved := '2secnutw3tchndagckluehlasti87l339p6czn93vnrelc83o52p0z7'; // write a sync structure, costs nothing, might help in the future
    dwVersion := 0;
    dwResultType := ertMAAResult;
  end;
end;

end.

